@@ -1,53 +1,70 @@
-// Copyright 2013-2018, The Gtk-rs Project Developers.
-// See the COPYRIGHT file at the top-level directory of this distribution.
-// Licensed under the MIT license, see the LICENSE file or <http://opensource.org/licenses/MIT>
+// Take a look at the license at the top of the repository in the LICENSE file.
 
-use glib_sys::{self, gboolean, gpointer};
+use crate::translate::{from_glib, from_glib_full, FromGlib, IntoGlib, ToGlibPtr};
+#[cfg(any(unix, feature = "dox"))]
+use crate::IOCondition;
+use ffi::{self, gboolean, gpointer};
 #[cfg(all(not(unix), feature = "dox"))]
 use libc::c_int as RawFd;
 use std::cell::RefCell;
 use std::mem::transmute;
+use std::num::NonZeroU32;
 #[cfg(unix)]
 use std::os::unix::io::RawFd;
-use std::process;
-use std::thread;
-use translate::{from_glib, from_glib_full, FromGlib, ToGlib, ToGlibPtr};
-#[cfg(any(unix, feature = "dox"))]
-use IOCondition;
+use std::time::Duration;
 
-use MainContext;
-use Source;
+use crate::MainContext;
+use crate::Source;
 
 /// The id of a source that is returned by `idle_add` and `timeout_add`.
+///
+/// This type does not implement `Clone` to prevent calling [`source_remove()`]
+/// multiple times on the same source.
 #[derive(Debug, Eq, PartialEq)]
-pub struct SourceId(u32);
+pub struct SourceId(NonZeroU32);
 
-#[doc(hidden)]
-impl ToGlib for SourceId {
-    type GlibType = u32;
-
-    #[inline]
-    fn to_glib(&self) -> u32 {
-        self.0
+impl SourceId {
+    /// Returns the internal source ID.
+    pub unsafe fn as_raw(&self) -> u32 {
+        self.0.get()
     }
 }
 
 #[doc(hidden)]
 impl FromGlib<u32> for SourceId {
     #[inline]
-    fn from_glib(val: u32) -> SourceId {
+    unsafe fn from_glib(val: u32) -> Self {
         assert_ne!(val, 0);
-        SourceId(val)
+        Self(NonZeroU32::new_unchecked(val))
     }
 }
 
 /// Process identificator
 #[derive(Copy, Clone, Debug, Eq, PartialEq)]
-pub struct Pid(pub glib_sys::GPid);
+#[doc(alias = "GPid")]
+pub struct Pid(pub ffi::GPid);
 
 unsafe impl Send for Pid {}
 unsafe impl Sync for Pid {}
 
+#[doc(hidden)]
+impl IntoGlib for Pid {
+    type GlibType = ffi::GPid;
+
+    #[inline]
+    fn into_glib(self) -> ffi::GPid {
+        self.0
+    }
+}
+
+#[doc(hidden)]
+impl FromGlib<ffi::GPid> for Pid {
+    #[inline]
+    unsafe fn from_glib(val: ffi::GPid) -> Self {
+        Self(val)
+    }
+}
+
 /// Continue calling the closure in the future iterations or drop it.
 ///
 /// This is the return type of `idle_add` and `timeout_add` closures.
@@ -59,50 +76,18 @@ unsafe impl Sync for Pid {}
 pub struct Continue(pub bool);
 
 #[doc(hidden)]
-impl ToGlib for Continue {
+impl IntoGlib for Continue {
     type GlibType = gboolean;
 
     #[inline]
-    fn to_glib(&self) -> gboolean {
-        self.0.to_glib()
-    }
-}
-
-/// Unwinding propagation guard. Aborts the process if destroyed while
-/// panicking.
-#[deprecated(note = "Rustc has this functionality built-in since 1.26.0")]
-pub struct CallbackGuard(());
-
-#[allow(deprecated)]
-impl CallbackGuard {
-    pub fn new() -> CallbackGuard {
-        CallbackGuard(())
-    }
-}
-
-#[allow(deprecated)]
-impl Default for CallbackGuard {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
-#[allow(deprecated)]
-impl Drop for CallbackGuard {
-    fn drop(&mut self) {
-        use std::io::stderr;
-        use std::io::Write;
-
-        if thread::panicking() {
-            let _ = stderr().write(b"Uncaught panic, exiting\n");
-            process::abort();
-        }
+    fn into_glib(self) -> gboolean {
+        self.0.into_glib()
     }
 }
 
 unsafe extern "C" fn trampoline<F: FnMut() -> Continue + 'static>(func: gpointer) -> gboolean {
     let func: &RefCell<F> = &*(func as *const RefCell<F>);
-    (&mut *func.borrow_mut())().to_glib()
+    (&mut *func.borrow_mut())().into_glib()
 }
 
 unsafe extern "C" fn destroy_closure<F: FnMut() -> Continue + 'static>(ptr: gpointer) {
@@ -115,7 +100,7 @@ fn into_raw<F: FnMut() -> Continue + 'static>(func: F) -> gpointer {
 }
 
 unsafe extern "C" fn trampoline_child_watch<F: FnMut(Pid, i32) + 'static>(
-    pid: glib_sys::GPid,
+    pid: ffi::GPid,
     status: i32,
     func: gpointer,
 ) {
@@ -133,16 +118,18 @@ fn into_raw_child_watch<F: FnMut(Pid, i32) + 'static>(func: F) -> gpointer {
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 unsafe extern "C" fn trampoline_unix_fd<F: FnMut(RawFd, IOCondition) -> Continue + 'static>(
     fd: i32,
-    condition: glib_sys::GIOCondition,
+    condition: ffi::GIOCondition,
     func: gpointer,
 ) -> gboolean {
     let func: &RefCell<F> = &*(func as *const RefCell<F>);
-    (&mut *func.borrow_mut())(fd, from_glib(condition)).to_glib()
+    (&mut *func.borrow_mut())(fd, from_glib(condition)).into_glib()
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 unsafe extern "C" fn destroy_closure_unix_fd<F: FnMut(RawFd, IOCondition) -> Continue + 'static>(
     ptr: gpointer,
 ) {
@@ -150,24 +137,65 @@ unsafe extern "C" fn destroy_closure_unix_fd<F: FnMut(RawFd, IOCondition) -> Con
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 fn into_raw_unix_fd<F: FnMut(RawFd, IOCondition) -> Continue + 'static>(func: F) -> gpointer {
     let func: Box<RefCell<F>> = Box::new(RefCell::new(func));
     Box::into_raw(func) as gpointer
 }
 
+/// Transform a generic FnOnce into a closure that can be used as callback in various glib methods
+///
+/// The resulting function can only be called once and will panic otherwise. It will return `Continue(false)`
+/// in order to prevent being called twice.
+#[inline(always)]
+fn fnmut_callback_wrapper(
+    func: impl FnOnce() + Send + 'static,
+) -> impl FnMut() -> Continue + Send + 'static {
+    let mut func = Some(func);
+    move || {
+        let func = func
+            .take()
+            .expect("GSource closure called after returning glib::Continue(false)");
+        func();
+        Continue(false)
+    }
+}
+
+/// Transform a generic FnOnce into a closure that can be used as callback in various glib methods
+///
+/// The resulting function can only be called once and will panic otherwise. It will return `Continue(false)`
+/// in order to prevent being called twice.
+///
+/// Different to `fnmut_callback_wrapper()`, this does not require `func` to be
+/// `Send` but can only be called from the thread that owns the main context.
+#[inline(always)]
+fn fnmut_callback_wrapper_local(
+    func: impl FnOnce() + 'static,
+) -> impl FnMut() -> Continue + 'static {
+    let mut func = Some(func);
+    move || {
+        let func = func
+            .take()
+            .expect("GSource closure called after returning glib::Continue(false)");
+        func();
+        Continue(false)
+    }
+}
+
 /// Adds a closure to be called by the default main loop when it's idle.
 ///
 /// `func` will be called repeatedly until it returns `Continue(false)`.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+#[doc(alias = "g_idle_add_full")]
 pub fn idle_add<F>(func: F) -> SourceId
 where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        from_glib(glib_sys::g_idle_add_full(
-            glib_sys::G_PRIORITY_DEFAULT_IDLE,
+        from_glib(ffi::g_idle_add_full(
+            ffi::G_PRIORITY_DEFAULT_IDLE,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
@@ -180,21 +208,39 @@ where
 /// `func` will be called repeatedly until it returns `Continue(false)`.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// In comparison to `idle_add()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_idle_add_full")]
+pub fn idle_add_once<F>(func: F) -> SourceId
+where
+    F: FnOnce() + Send + 'static,
+{
+    idle_add(fnmut_callback_wrapper(func))
+}
+
+/// Adds a closure to be called by the default main loop when it's idle.
+///
+/// `func` will be called repeatedly until it returns `Continue(false)`.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
 ///
 /// Different to `idle_add()`, this does not require `func` to be
 /// `Send` but can only be called from the thread that owns the main context.
 ///
 /// This function panics if called from a different thread than the one that
 /// owns the main context.
+#[doc(alias = "g_idle_add_full")]
 pub fn idle_add_local<F>(func: F) -> SourceId
 where
     F: FnMut() -> Continue + 'static,
 {
     unsafe {
         assert!(MainContext::default().is_owner());
-        from_glib(glib_sys::g_idle_add_full(
-            glib_sys::G_PRIORITY_DEFAULT_IDLE,
+        from_glib(ffi::g_idle_add_full(
+            ffi::G_PRIORITY_DEFAULT_IDLE,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
@@ -202,6 +248,29 @@ where
     }
 }
 
+/// Adds a closure to be called by the default main loop when it's idle.
+///
+/// `func` will be called repeatedly until it returns `Continue(false)`.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// Different to `idle_add()`, this does not require `func` to be
+/// `Send` but can only be called from the thread that owns the main context.
+///
+/// This function panics if called from a different thread than the one that
+/// owns the main context.
+///
+/// In comparison to `idle_add_local()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_idle_add_full")]
+pub fn idle_add_local_once<F>(func: F) -> SourceId
+where
+    F: FnOnce() + 'static,
+{
+    idle_add_local(fnmut_callback_wrapper_local(func))
+}
+
 /// Adds a closure to be called by the default main loop at regular intervals
 /// with millisecond granularity.
 ///
@@ -211,15 +280,16 @@ where
 /// precision is not necessary.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
-pub fn timeout_add<F>(interval: u32, func: F) -> SourceId
+/// Thus, the closure is called on the main thread.
+#[doc(alias = "g_timeout_add_full")]
+pub fn timeout_add<F>(interval: Duration, func: F) -> SourceId
 where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        from_glib(glib_sys::g_timeout_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
-            interval,
+        from_glib(ffi::g_timeout_add_full(
+            ffi::G_PRIORITY_DEFAULT,
+            interval.as_millis() as _,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
@@ -236,22 +306,44 @@ where
 /// precision is not necessary.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// In comparison to `timeout_add()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_timeout_add_full")]
+pub fn timeout_add_once<F>(interval: Duration, func: F) -> SourceId
+where
+    F: FnOnce() + Send + 'static,
+{
+    timeout_add(interval, fnmut_callback_wrapper(func))
+}
+
+/// Adds a closure to be called by the default main loop at regular intervals
+/// with millisecond granularity.
+///
+/// `func` will be called repeatedly every `interval` milliseconds until it
+/// returns `Continue(false)`. Precise timing is not guaranteed, the timeout may
+/// be delayed by other events. Prefer `timeout_add_seconds` when millisecond
+/// precision is not necessary.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
 ///
 /// Different to `timeout_add()`, this does not require `func` to be
 /// `Send` but can only be called from the thread that owns the main context.
 ///
 /// This function panics if called from a different thread than the one that
 /// owns the main context.
-pub fn timeout_add_local<F>(interval: u32, func: F) -> SourceId
+#[doc(alias = "g_timeout_add_full")]
+pub fn timeout_add_local<F>(interval: Duration, func: F) -> SourceId
 where
     F: FnMut() -> Continue + 'static,
 {
     unsafe {
         assert!(MainContext::default().is_owner());
-        from_glib(glib_sys::g_timeout_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
-            interval,
+        from_glib(ffi::g_timeout_add_full(
+            ffi::G_PRIORITY_DEFAULT,
+            interval.as_millis() as _,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
@@ -259,6 +351,33 @@ where
     }
 }
 
+/// Adds a closure to be called by the default main loop at regular intervals
+/// with millisecond granularity.
+///
+/// `func` will be called repeatedly every `interval` milliseconds until it
+/// returns `Continue(false)`. Precise timing is not guaranteed, the timeout may
+/// be delayed by other events. Prefer `timeout_add_seconds` when millisecond
+/// precision is not necessary.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// Different to `timeout_add()`, this does not require `func` to be
+/// `Send` but can only be called from the thread that owns the main context.
+///
+/// This function panics if called from a different thread than the one that
+/// owns the main context.
+///
+/// In comparison to `timeout_add_local()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_timeout_add_full")]
+pub fn timeout_add_local_once<F>(interval: Duration, func: F) -> SourceId
+where
+    F: FnOnce() + 'static,
+{
+    timeout_add_local(interval, fnmut_callback_wrapper_local(func))
+}
+
 /// Adds a closure to be called by the default main loop at regular intervals
 /// with second granularity.
 ///
@@ -267,14 +386,15 @@ where
 /// be delayed by other events.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+#[doc(alias = "g_timeout_add_seconds_full")]
 pub fn timeout_add_seconds<F>(interval: u32, func: F) -> SourceId
 where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        from_glib(glib_sys::g_timeout_add_seconds_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_timeout_add_seconds_full(
+            ffi::G_PRIORITY_DEFAULT,
             interval,
             Some(trampoline::<F>),
             into_raw(func),
@@ -291,21 +411,42 @@ where
 /// be delayed by other events.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// In comparison to `timeout_add_seconds()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_timeout_add_seconds_full")]
+pub fn timeout_add_seconds_once<F>(interval: u32, func: F) -> SourceId
+where
+    F: FnOnce() + Send + 'static,
+{
+    timeout_add_seconds(interval, fnmut_callback_wrapper(func))
+}
+
+/// Adds a closure to be called by the default main loop at regular intervals
+/// with second granularity.
+///
+/// `func` will be called repeatedly every `interval` seconds until it
+/// returns `Continue(false)`. Precise timing is not guaranteed, the timeout may
+/// be delayed by other events.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
 ///
 /// Different to `timeout_add_seconds()`, this does not require `func` to be
 /// `Send` but can only be called from the thread that owns the main context.
 ///
 /// This function panics if called from a different thread than the one that
 /// owns the main context.
+#[doc(alias = "g_timeout_add_seconds_full")]
 pub fn timeout_add_seconds_local<F>(interval: u32, func: F) -> SourceId
 where
     F: FnMut() -> Continue + 'static,
 {
     unsafe {
         assert!(MainContext::default().is_owner());
-        from_glib(glib_sys::g_timeout_add_seconds_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_timeout_add_seconds_full(
+            ffi::G_PRIORITY_DEFAULT,
             interval,
             Some(trampoline::<F>),
             into_raw(func),
@@ -314,19 +455,46 @@ where
     }
 }
 
+/// Adds a closure to be called by the default main loop at regular intervals
+/// with second granularity.
+///
+/// `func` will be called repeatedly every `interval` seconds until it
+/// returns `Continue(false)`. Precise timing is not guaranteed, the timeout may
+/// be delayed by other events.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// Different to `timeout_add_seconds()`, this does not require `func` to be
+/// `Send` but can only be called from the thread that owns the main context.
+///
+/// This function panics if called from a different thread than the one that
+/// owns the main context.
+///
+/// In comparison to `timeout_add_seconds_local()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_timeout_add_seconds_full")]
+pub fn timeout_add_seconds_local_once<F>(interval: u32, func: F) -> SourceId
+where
+    F: FnOnce() + 'static,
+{
+    timeout_add_seconds_local(interval, fnmut_callback_wrapper_local(func))
+}
+
 /// Adds a closure to be called by the main loop the returned `Source` is attached to when a child
 /// process exits.
 ///
 /// `func` will be called when `pid` exits
+#[doc(alias = "g_child_watch_add_full")]
 pub fn child_watch_add<F>(pid: Pid, func: F) -> SourceId
 where
     F: FnMut(Pid, i32) + Send + 'static,
 {
     unsafe {
-        from_glib(glib_sys::g_child_watch_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_child_watch_add_full(
+            ffi::G_PRIORITY_DEFAULT,
             pid.0,
-            Some(transmute(trampoline_child_watch::<F> as usize)),
+            Some(trampoline_child_watch::<F>),
             into_raw_child_watch(func),
             Some(destroy_closure_child_watch::<F>),
         ))
@@ -343,16 +511,17 @@ where
 ///
 /// This function panics if called from a different thread than the one that
 /// owns the main context.
+#[doc(alias = "g_child_watch_add_full")]
 pub fn child_watch_add_local<F>(pid: Pid, func: F) -> SourceId
 where
     F: FnMut(Pid, i32) + 'static,
 {
     unsafe {
         assert!(MainContext::default().is_owner());
-        from_glib(glib_sys::g_child_watch_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_child_watch_add_full(
+            ffi::G_PRIORITY_DEFAULT,
             pid.0,
-            Some(transmute(trampoline_child_watch::<F> as usize)),
+            Some(trampoline_child_watch::<F>),
             into_raw_child_watch(func),
             Some(destroy_closure_child_watch::<F>),
         ))
@@ -360,20 +529,22 @@ where
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 /// Adds a closure to be called by the default main loop whenever a UNIX signal is raised.
 ///
 /// `func` will be called repeatedly every time `signum` is raised until it
 /// returns `Continue(false)`.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+#[doc(alias = "g_unix_signal_add_full")]
 pub fn unix_signal_add<F>(signum: i32, func: F) -> SourceId
 where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        from_glib(glib_sys::g_unix_signal_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_unix_signal_add_full(
+            ffi::G_PRIORITY_DEFAULT,
             signum,
             Some(trampoline::<F>),
             into_raw(func),
@@ -383,27 +554,49 @@ where
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 /// Adds a closure to be called by the default main loop whenever a UNIX signal is raised.
 ///
 /// `func` will be called repeatedly every time `signum` is raised until it
 /// returns `Continue(false)`.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// In comparison to `unix_signal_add()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_unix_signal_add_full")]
+pub fn unix_signal_add_once<F>(signum: i32, func: F) -> SourceId
+where
+    F: FnOnce() + Send + 'static,
+{
+    unix_signal_add(signum, fnmut_callback_wrapper(func))
+}
+
+#[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
+/// Adds a closure to be called by the default main loop whenever a UNIX signal is raised.
+///
+/// `func` will be called repeatedly every time `signum` is raised until it
+/// returns `Continue(false)`.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
 ///
 /// Different to `unix_signal_add()`, this does not require `func` to be
 /// `Send` but can only be called from the thread that owns the main context.
 ///
 /// This function panics if called from a different thread than the one that
 /// owns the main context.
+#[doc(alias = "g_unix_signal_add_full")]
 pub fn unix_signal_add_local<F>(signum: i32, func: F) -> SourceId
 where
     F: FnMut() -> Continue + 'static,
 {
     unsafe {
         assert!(MainContext::default().is_owner());
-        from_glib(glib_sys::g_unix_signal_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_unix_signal_add_full(
+            ffi::G_PRIORITY_DEFAULT,
             signum,
             Some(trampoline::<F>),
             into_raw(func),
@@ -413,6 +606,33 @@ where
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
+/// Adds a closure to be called by the default main loop whenever a UNIX signal is raised.
+///
+/// `func` will be called repeatedly every time `signum` is raised until it
+/// returns `Continue(false)`.
+///
+/// The default main loop almost always is the main loop of the main thread.
+/// Thus, the closure is called on the main thread.
+///
+/// Different to `unix_signal_add()`, this does not require `func` to be
+/// `Send` but can only be called from the thread that owns the main context.
+///
+/// This function panics if called from a different thread than the one that
+/// owns the main context.
+///
+/// In comparison to `unix_signal_add_local()`, this only requires `func` to be
+/// `FnOnce`, and will automatically return `Continue(false)`.
+#[doc(alias = "g_unix_signal_add_full")]
+pub fn unix_signal_add_local_once<F>(signum: i32, func: F) -> SourceId
+where
+    F: FnOnce() + 'static,
+{
+    unix_signal_add_local(signum, fnmut_callback_wrapper_local(func))
+}
+
+#[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 /// Adds a closure to be called by the main loop the returned `Source` is attached to whenever a
 /// UNIX file descriptor reaches the given IO condition.
 ///
@@ -420,17 +640,18 @@ where
 /// until it returns `Continue(false)`.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
+#[doc(alias = "g_unix_fd_add_full")]
 pub fn unix_fd_add<F>(fd: RawFd, condition: IOCondition, func: F) -> SourceId
 where
     F: FnMut(RawFd, IOCondition) -> Continue + Send + 'static,
 {
     unsafe {
-        from_glib(glib_sys::g_unix_fd_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_unix_fd_add_full(
+            ffi::G_PRIORITY_DEFAULT,
             fd,
-            condition.to_glib(),
-            Some(transmute(trampoline_unix_fd::<F> as usize)),
+            condition.into_glib(),
+            Some(trampoline_unix_fd::<F>),
             into_raw_unix_fd(func),
             Some(destroy_closure_unix_fd::<F>),
         ))
@@ -438,6 +659,7 @@ where
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 /// Adds a closure to be called by the main loop the returned `Source` is attached to whenever a
 /// UNIX file descriptor reaches the given IO condition.
 ///
@@ -445,24 +667,25 @@ where
 /// until it returns `Continue(false)`.
 ///
 /// The default main loop almost always is the main loop of the main thread.
-/// Thus the closure is called on the main thread.
+/// Thus, the closure is called on the main thread.
 ///
 /// Different to `unix_fd_add()`, this does not require `func` to be
 /// `Send` but can only be called from the thread that owns the main context.
 ///
 /// This function panics if called from a different thread than the one that
 /// owns the main context.
+#[doc(alias = "g_unix_fd_add_full")]
 pub fn unix_fd_add_local<F>(fd: RawFd, condition: IOCondition, func: F) -> SourceId
 where
     F: FnMut(RawFd, IOCondition) -> Continue + 'static,
 {
     unsafe {
         assert!(MainContext::default().is_owner());
-        from_glib(glib_sys::g_unix_fd_add_full(
-            glib_sys::G_PRIORITY_DEFAULT,
+        from_glib(ffi::g_unix_fd_add_full(
+            ffi::G_PRIORITY_DEFAULT,
             fd,
-            condition.to_glib(),
-            Some(transmute(trampoline_unix_fd::<F> as usize)),
+            condition.into_glib(),
+            Some(trampoline_unix_fd::<F>),
             into_raw_unix_fd(func),
             Some(destroy_closure_unix_fd::<F>),
         ))
@@ -477,9 +700,10 @@ where
 /// For historical reasons, the native function always returns true, so we
 /// ignore it here.
 #[allow(clippy::needless_pass_by_value)]
+#[doc(alias = "g_source_remove")]
 pub fn source_remove(source_id: SourceId) {
     unsafe {
-        glib_sys::g_source_remove(source_id.to_glib());
+        ffi::g_source_remove(source_id.as_raw());
     }
 }
 
@@ -489,11 +713,11 @@ pub fn source_remove(source_id: SourceId) {
 pub struct Priority(i32);
 
 #[doc(hidden)]
-impl ToGlib for Priority {
+impl IntoGlib for Priority {
     type GlibType = i32;
 
     #[inline]
-    fn to_glib(&self) -> i32 {
+    fn into_glib(self) -> i32 {
         self.0
     }
 }
@@ -501,42 +725,43 @@ impl ToGlib for Priority {
 #[doc(hidden)]
 impl FromGlib<i32> for Priority {
     #[inline]
-    fn from_glib(val: i32) -> Priority {
-        Priority(val)
+    unsafe fn from_glib(val: i32) -> Self {
+        Self(val)
     }
 }
 
 impl Default for Priority {
-    fn default() -> Priority {
+    fn default() -> Self {
         PRIORITY_DEFAULT
     }
 }
 
-pub const PRIORITY_HIGH: Priority = Priority(glib_sys::G_PRIORITY_HIGH);
-pub const PRIORITY_DEFAULT: Priority = Priority(glib_sys::G_PRIORITY_DEFAULT);
-pub const PRIORITY_HIGH_IDLE: Priority = Priority(glib_sys::G_PRIORITY_HIGH_IDLE);
-pub const PRIORITY_DEFAULT_IDLE: Priority = Priority(glib_sys::G_PRIORITY_DEFAULT_IDLE);
-pub const PRIORITY_LOW: Priority = Priority(glib_sys::G_PRIORITY_LOW);
+pub const PRIORITY_HIGH: Priority = Priority(ffi::G_PRIORITY_HIGH);
+pub const PRIORITY_DEFAULT: Priority = Priority(ffi::G_PRIORITY_DEFAULT);
+pub const PRIORITY_HIGH_IDLE: Priority = Priority(ffi::G_PRIORITY_HIGH_IDLE);
+pub const PRIORITY_DEFAULT_IDLE: Priority = Priority(ffi::G_PRIORITY_DEFAULT_IDLE);
+pub const PRIORITY_LOW: Priority = Priority(ffi::G_PRIORITY_LOW);
 
 /// Adds a closure to be called by the main loop the return `Source` is attached to when it's idle.
 ///
 /// `func` will be called repeatedly until it returns `Continue(false)`.
+#[doc(alias = "g_idle_source_new")]
 pub fn idle_source_new<F>(name: Option<&str>, priority: Priority, func: F) -> Source
 where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        let source = glib_sys::g_idle_source_new();
-        glib_sys::g_source_set_callback(
+        let source = ffi::g_idle_source_new();
+        ffi::g_source_set_callback(
             source,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
         );
-        glib_sys::g_source_set_priority(source, priority.to_glib());
+        ffi::g_source_set_priority(source, priority.into_glib());
 
         if let Some(name) = name {
-            glib_sys::g_source_set_name(source, name.to_glib_none().0);
+            ffi::g_source_set_name(source, name.to_glib_none().0);
         }
 
         from_glib_full(source)
@@ -550,8 +775,9 @@ where
 /// returns `Continue(false)`. Precise timing is not guaranteed, the timeout may
 /// be delayed by other events. Prefer `timeout_add_seconds` when millisecond
 /// precision is not necessary.
+#[doc(alias = "g_timeout_source_new")]
 pub fn timeout_source_new<F>(
-    interval: u32,
+    interval: Duration,
     name: Option<&str>,
     priority: Priority,
     func: F,
@@ -560,17 +786,17 @@ where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        let source = glib_sys::g_timeout_source_new(interval);
-        glib_sys::g_source_set_callback(
+        let source = ffi::g_timeout_source_new(interval.as_millis() as _);
+        ffi::g_source_set_callback(
             source,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
         );
-        glib_sys::g_source_set_priority(source, priority.to_glib());
+        ffi::g_source_set_priority(source, priority.into_glib());
 
         if let Some(name) = name {
-            glib_sys::g_source_set_name(source, name.to_glib_none().0);
+            ffi::g_source_set_name(source, name.to_glib_none().0);
         }
 
         from_glib_full(source)
@@ -583,6 +809,7 @@ where
 /// `func` will be called repeatedly every `interval` seconds until it
 /// returns `Continue(false)`. Precise timing is not guaranteed, the timeout may
 /// be delayed by other events.
+#[doc(alias = "g_timeout_source_new_seconds")]
 pub fn timeout_source_new_seconds<F>(
     interval: u32,
     name: Option<&str>,
@@ -593,17 +820,17 @@ where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        let source = glib_sys::g_timeout_source_new_seconds(interval);
-        glib_sys::g_source_set_callback(
+        let source = ffi::g_timeout_source_new_seconds(interval);
+        ffi::g_source_set_callback(
             source,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
         );
-        glib_sys::g_source_set_priority(source, priority.to_glib());
+        ffi::g_source_set_priority(source, priority.into_glib());
 
         if let Some(name) = name {
-            glib_sys::g_source_set_name(source, name.to_glib_none().0);
+            ffi::g_source_set_name(source, name.to_glib_none().0);
         }
 
         from_glib_full(source)
@@ -614,6 +841,7 @@ where
 /// process exits.
 ///
 /// `func` will be called when `pid` exits
+#[doc(alias = "g_child_watch_source_new")]
 pub fn child_watch_source_new<F>(
     pid: Pid,
     name: Option<&str>,
@@ -624,17 +852,20 @@ where
     F: FnMut(Pid, i32) + Send + 'static,
 {
     unsafe {
-        let source = glib_sys::g_child_watch_source_new(pid.0);
-        glib_sys::g_source_set_callback(
+        let source = ffi::g_child_watch_source_new(pid.0);
+        ffi::g_source_set_callback(
             source,
-            Some(transmute(trampoline_child_watch::<F> as usize)),
+            Some(transmute::<
+                _,
+                unsafe extern "C" fn(ffi::gpointer) -> ffi::gboolean,
+            >(trampoline_child_watch::<F> as *const ())),
             into_raw_child_watch(func),
             Some(destroy_closure_child_watch::<F>),
         );
-        glib_sys::g_source_set_priority(source, priority.to_glib());
+        ffi::g_source_set_priority(source, priority.into_glib());
 
         if let Some(name) = name {
-            glib_sys::g_source_set_name(source, name.to_glib_none().0);
+            ffi::g_source_set_name(source, name.to_glib_none().0);
         }
 
         from_glib_full(source)
@@ -642,11 +873,13 @@ where
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 /// Adds a closure to be called by the main loop the returned `Source` is attached to whenever a
 /// UNIX signal is raised.
 ///
 /// `func` will be called repeatedly every time `signum` is raised until it
 /// returns `Continue(false)`.
+#[doc(alias = "g_unix_signal_source_new")]
 pub fn unix_signal_source_new<F>(
     signum: i32,
     name: Option<&str>,
@@ -657,17 +890,17 @@ where
     F: FnMut() -> Continue + Send + 'static,
 {
     unsafe {
-        let source = glib_sys::g_unix_signal_source_new(signum);
-        glib_sys::g_source_set_callback(
+        let source = ffi::g_unix_signal_source_new(signum);
+        ffi::g_source_set_callback(
             source,
             Some(trampoline::<F>),
             into_raw(func),
             Some(destroy_closure::<F>),
         );
-        glib_sys::g_source_set_priority(source, priority.to_glib());
+        ffi::g_source_set_priority(source, priority.into_glib());
 
         if let Some(name) = name {
-            glib_sys::g_source_set_name(source, name.to_glib_none().0);
+            ffi::g_source_set_name(source, name.to_glib_none().0);
         }
 
         from_glib_full(source)
@@ -675,11 +908,13 @@ where
 }
 
 #[cfg(any(unix, feature = "dox"))]
+#[cfg_attr(feature = "dox", doc(cfg(unix)))]
 /// Adds a closure to be called by the main loop the returned `Source` is attached to whenever a
 /// UNIX file descriptor reaches the given IO condition.
 ///
 /// `func` will be called repeatedly while the file descriptor matches the given IO condition
 /// until it returns `Continue(false)`.
+#[doc(alias = "g_unix_fd_source_new")]
 pub fn unix_fd_source_new<F>(
     fd: RawFd,
     condition: IOCondition,
@@ -691,17 +926,20 @@ where
     F: FnMut(RawFd, IOCondition) -> Continue + Send + 'static,
 {
     unsafe {
-        let source = glib_sys::g_unix_fd_source_new(fd, condition.to_glib());
-        glib_sys::g_source_set_callback(
+        let source = ffi::g_unix_fd_source_new(fd, condition.into_glib());
+        ffi::g_source_set_callback(
             source,
-            Some(transmute(trampoline_unix_fd::<F> as usize)),
+            Some(transmute::<
+                _,
+                unsafe extern "C" fn(ffi::gpointer) -> ffi::gboolean,
+            >(trampoline_unix_fd::<F> as *const ())),
             into_raw_unix_fd(func),
             Some(destroy_closure_unix_fd::<F>),
         );
-        glib_sys::g_source_set_priority(source, priority.to_glib());
+        ffi::g_source_set_priority(source, priority.into_glib());
 
         if let Some(name) = name {
-            glib_sys::g_source_set_name(source, name.to_glib_none().0);
+            ffi::g_source_set_name(source, name.to_glib_none().0);
         }
 
         from_glib_full(source)
@@ -709,19 +947,21 @@ where
 }
 
 impl Source {
+    #[doc(alias = "g_source_attach")]
     pub fn attach(&self, context: Option<&MainContext>) -> SourceId {
         unsafe {
-            from_glib(glib_sys::g_source_attach(
+            from_glib(ffi::g_source_attach(
                 self.to_glib_none().0,
                 context.to_glib_none().0,
             ))
         }
     }
 
-    pub fn remove(tag: SourceId) -> Result<(), ::BoolError> {
+    #[doc(alias = "g_source_remove")]
+    pub fn remove(tag: SourceId) -> Result<(), crate::BoolError> {
         unsafe {
-            glib_result_from_gboolean!(
-                glib_sys::g_source_remove(tag.to_glib()),
+            result_from_gboolean!(
+                ffi::g_source_remove(tag.as_raw()),
                 "Failed to remove source"
             )
         }
