@@ -1,19 +1,15 @@
-// Copyright 2013-2016, The Gtk-rs Project Developers.
-// See the COPYRIGHT file at the top-level directory of this distribution.
-// Licensed under the MIT license, see the LICENSE file or <http://opensource.org/licenses/MIT>
+// Take a look at the license at the top of the repository in the LICENSE file.
 
-use glib_sys;
-use gobject_sys;
+use crate::translate::*;
+use crate::types::StaticType;
+use crate::types::Type;
+use crate::BoolError;
 use std::borrow::{Borrow, Cow, ToOwned};
 use std::cmp::{Eq, PartialEq};
 use std::fmt;
 use std::hash::{Hash, Hasher};
 use std::ops::Deref;
 use std::slice;
-use translate::*;
-use types::StaticType;
-use types::Type;
-use value::{FromValueOptional, SetValue, SetValueOptional, Value};
 
 /// Describes `Variant` types.
 ///
@@ -21,10 +17,11 @@ use value::{FromValueOptional, SetValue, SetValueOptional, Value};
 /// "type strings". `VariantType` is an owned immutable type string (you can
 /// think of it as a `Box<str>` statically guaranteed to be a valid type
 /// string), `&VariantTy` is a borrowed one (like `&str`).
+#[doc(alias = "GVariantType")]
 pub struct VariantType {
     // GVariantType* essentially is a char*, that always is valid UTF-8 but
     // isn't NUL-terminated.
-    ptr: *mut glib_sys::GVariantType,
+    ptr: *mut ffi::GVariantType,
     // We query the length on creation assuming it's cheap (because type strings
     // are short) and likely to happen anyway.
     len: usize,
@@ -34,7 +31,7 @@ impl VariantType {
     /// Tries to create a `VariantType` from a string slice.
     ///
     /// Returns `Ok` if the string is a valid type string, `Err` otherwise.
-    pub fn new(type_string: &str) -> Result<VariantType, ()> {
+    pub fn new(type_string: &str) -> Result<VariantType, BoolError> {
         VariantTy::new(type_string).map(ToOwned::to_owned)
     }
 }
@@ -44,7 +41,7 @@ unsafe impl Sync for VariantType {}
 
 impl Drop for VariantType {
     fn drop(&mut self) {
-        unsafe { glib_sys::g_variant_type_free(self.ptr) }
+        unsafe { ffi::g_variant_type_free(self.ptr) }
     }
 }
 
@@ -58,7 +55,7 @@ impl Clone for VariantType {
     fn clone(&self) -> VariantType {
         unsafe {
             VariantType {
-                ptr: glib_sys::g_variant_type_copy(self.ptr),
+                ptr: ffi::g_variant_type_copy(self.ptr),
                 len: self.len,
             }
         }
@@ -93,31 +90,57 @@ impl Hash for VariantType {
     }
 }
 
-impl<'a> Into<Cow<'a, VariantTy>> for VariantType {
-    fn into(self) -> Cow<'a, VariantTy> {
-        Cow::Owned(self)
+impl<'a> From<VariantType> for Cow<'a, VariantTy> {
+    fn from(ty: VariantType) -> Cow<'a, VariantTy> {
+        Cow::Owned(ty)
     }
 }
 
 #[doc(hidden)]
-impl<'a> ToGlibPtrMut<'a, *mut glib_sys::GVariantType> for VariantType {
+impl<'a> ToGlibPtr<'a, *const ffi::GVariantType> for VariantType {
+    type Storage = &'a Self;
+
+    fn to_glib_none(&'a self) -> Stash<'a, *const ffi::GVariantType, Self> {
+        Stash(self.ptr, self)
+    }
+
+    fn to_glib_full(&self) -> *const ffi::GVariantType {
+        unsafe { ffi::g_variant_type_copy(self.ptr) }
+    }
+}
+
+#[doc(hidden)]
+impl<'a> ToGlibPtr<'a, *mut ffi::GVariantType> for VariantType {
+    type Storage = &'a Self;
+
+    fn to_glib_none(&'a self) -> Stash<'a, *mut ffi::GVariantType, Self> {
+        Stash(self.ptr, self)
+    }
+
+    fn to_glib_full(&self) -> *mut ffi::GVariantType {
+        unsafe { ffi::g_variant_type_copy(self.ptr) }
+    }
+}
+
+#[doc(hidden)]
+impl<'a> ToGlibPtrMut<'a, *mut ffi::GVariantType> for VariantType {
     type Storage = &'a mut Self;
 
-    fn to_glib_none_mut(&'a mut self) -> StashMut<'a, *mut glib_sys::GVariantType, Self> {
+    fn to_glib_none_mut(&'a mut self) -> StashMut<'a, *mut ffi::GVariantType, Self> {
         StashMut(self.ptr, self)
     }
 }
 
 #[doc(hidden)]
-impl FromGlibPtrNone<*const glib_sys::GVariantType> for VariantType {
-    unsafe fn from_glib_none(ptr: *const glib_sys::GVariantType) -> VariantType {
+impl FromGlibPtrNone<*const ffi::GVariantType> for VariantType {
+    unsafe fn from_glib_none(ptr: *const ffi::GVariantType) -> VariantType {
         VariantTy::from_ptr(ptr).to_owned()
     }
 }
 
 #[doc(hidden)]
-impl FromGlibPtrFull<*const glib_sys::GVariantType> for VariantType {
-    unsafe fn from_glib_full(ptr: *const glib_sys::GVariantType) -> VariantType {
+impl FromGlibPtrFull<*const ffi::GVariantType> for VariantType {
+    unsafe fn from_glib_full(ptr: *const ffi::GVariantType) -> VariantType {
         // Don't assume ownership of a const pointer.
         // A transfer: full annotation on a `const GVariantType*` is likely a bug.
         VariantTy::from_ptr(ptr).to_owned()
@@ -134,15 +157,19 @@ pub struct VariantTy {
 }
 
 impl VariantTy {
+    /// An indefinite type that is a supertype of every type (including itself).
+    #[doc(alias = "G_VARIANT_TYPE_ANY")]
+    pub const ANY: &'static VariantTy = unsafe { std::mem::transmute::<&str, &VariantTy>("*") };
+
     /// Tries to create a `&VariantTy` from a string slice.
     ///
     /// Returns `Ok` if the string is a valid type string, `Err` otherwise.
-    pub fn new(type_string: &str) -> Result<&VariantTy, ()> {
+    pub fn new(type_string: &str) -> Result<&VariantTy, BoolError> {
         let ptr = type_string.as_ptr();
         let limit = ptr as usize + type_string.len();
         let mut end = 0_usize;
         unsafe {
-            let ok = from_glib(glib_sys::g_variant_type_string_scan(
+            let ok = from_glib(ffi::g_variant_type_string_scan(
                 ptr as *const _,
                 limit as *const _,
                 &mut end as *mut usize as *mut _,
@@ -150,12 +177,17 @@ impl VariantTy {
             if ok && end == limit {
                 Ok(&*(type_string.as_bytes() as *const [u8] as *const VariantTy))
             } else {
-                Err(())
+                Err(bool_error!("Invalid type string: '{}'", type_string))
             }
         }
     }
 
     /// Converts a type string into `&VariantTy` without any checks.
+    ///
+    /// # Safety
+    ///
+    /// The caller is responsible for passing in only a valid variant type string
+    /// which is already registered with the type system.
     pub unsafe fn from_str_unchecked(type_string: &str) -> &VariantTy {
         &*(type_string as *const str as *const VariantTy)
     }
@@ -163,14 +195,14 @@ impl VariantTy {
     /// Creates `&VariantTy` with a wildcard lifetime from a `GVariantType`
     /// pointer.
     #[doc(hidden)]
-    pub unsafe fn from_ptr<'a>(ptr: *const glib_sys::GVariantType) -> &'a VariantTy {
-        let len = glib_sys::g_variant_type_get_string_length(ptr) as usize;
+    pub unsafe fn from_ptr<'a>(ptr: *const ffi::GVariantType) -> &'a VariantTy {
+        let len = ffi::g_variant_type_get_string_length(ptr) as usize;
         &*(slice::from_raw_parts(ptr as *const u8, len) as *const [u8] as *const VariantTy)
     }
 
     /// Returns a `GVariantType` pointer.
     #[doc(hidden)]
-    pub fn as_ptr(&self) -> *const glib_sys::GVariantType {
+    pub fn as_ptr(&self) -> *const ffi::GVariantType {
         self.inner.as_ptr() as *const _
     }
 
@@ -178,15 +210,20 @@ impl VariantTy {
     pub fn to_str(&self) -> &str {
         &self.inner
     }
+
+    /// Return this type as an array.
+    pub(crate) fn with_array(&self) -> VariantType {
+        VariantType::new(&format!("a{}", self.to_str())).expect("invalid variant signature")
+    }
 }
 
 unsafe impl Sync for VariantTy {}
 
 #[doc(hidden)]
-impl<'a> ToGlibPtr<'a, *const glib_sys::GVariantType> for VariantTy {
+impl<'a> ToGlibPtr<'a, *const ffi::GVariantType> for VariantTy {
     type Storage = &'a Self;
 
-    fn to_glib_none(&'a self) -> Stash<'a, *const glib_sys::GVariantType, Self> {
+    fn to_glib_none(&'a self) -> Stash<'a, *const ffi::GVariantType, Self> {
         Stash(self.as_ptr(), self)
     }
 }
@@ -197,9 +234,9 @@ impl fmt::Display for VariantTy {
     }
 }
 
-impl<'a> Into<Cow<'a, VariantTy>> for &'a VariantTy {
-    fn into(self) -> Cow<'a, VariantTy> {
-        Cow::Borrowed(self)
+impl<'a> From<&'a VariantTy> for Cow<'a, VariantTy> {
+    fn from(ty: &'a VariantTy) -> Cow<'a, VariantTy> {
+        Cow::Borrowed(ty)
     }
 }
 
@@ -209,7 +246,7 @@ impl ToOwned for VariantTy {
     fn to_owned(&self) -> VariantType {
         unsafe {
             VariantType {
-                ptr: glib_sys::g_variant_type_copy(self.as_ptr()),
+                ptr: ffi::g_variant_type_copy(self.as_ptr()),
                 len: self.inner.len(),
             }
         }
@@ -218,73 +255,117 @@ impl ToOwned for VariantTy {
 
 impl StaticType for VariantTy {
     fn static_type() -> Type {
-        unsafe { from_glib(glib_sys::g_variant_type_get_gtype()) }
+        unsafe { from_glib(ffi::g_variant_type_get_gtype()) }
     }
 }
 
-impl SetValue for VariantTy {
-    unsafe fn set_value(value: &mut Value, this: &Self) {
-        gobject_sys::g_value_set_boxed(
-            value.to_glib_none_mut().0,
-            this.to_glib_none().0 as glib_sys::gpointer,
-        )
+#[doc(hidden)]
+unsafe impl<'a> crate::value::FromValue<'a> for &'a VariantTy {
+    type Checker = crate::value::GenericValueTypeOrNoneChecker<Self>;
+
+    unsafe fn from_value(value: &'a crate::Value) -> Self {
+        let ptr = gobject_ffi::g_value_get_boxed(value.to_glib_none().0);
+        assert!(!ptr.is_null());
+        VariantTy::from_ptr(ptr as *const ffi::GVariantType)
     }
 }
 
-impl SetValueOptional for VariantTy {
-    unsafe fn set_value_optional(value: &mut Value, this: Option<&Self>) {
-        use std::ptr;
-        let p = match this {
-            Some(ref t) => t.to_glib_none().0 as glib_sys::gpointer,
-            None => ptr::null(),
-        };
-        gobject_sys::g_value_set_boxed(value.to_glib_none_mut().0, p)
-    }
-}
-
-impl<'a> FromValueOptional<'a> for &'a VariantTy {
-    unsafe fn from_value_optional(value: &'a Value) -> Option<Self> {
-        let cvty =
-            gobject_sys::g_value_get_boxed(value.to_glib_none().0) as *mut glib_sys::GVariantType;
-        if cvty.is_null() {
-            None
-        } else {
-            Some(VariantTy::from_ptr(cvty))
+#[doc(hidden)]
+impl crate::value::ToValue for VariantTy {
+    fn to_value(&self) -> crate::Value {
+        unsafe {
+            let mut value = crate::Value::from_type(VariantTy::static_type());
+            gobject_ffi::g_value_set_boxed(
+                value.to_glib_none_mut().0,
+                self.to_glib_none().0 as *mut _,
+            );
+            value
         }
     }
+
+    fn value_type(&self) -> crate::Type {
+        VariantTy::static_type()
+    }
+}
+
+#[doc(hidden)]
+impl crate::value::ToValue for &VariantTy {
+    fn to_value(&self) -> crate::Value {
+        (*self).to_value()
+    }
+
+    fn value_type(&self) -> crate::Type {
+        VariantTy::static_type()
+    }
+}
+
+#[doc(hidden)]
+impl crate::value::ToValueOptional for &VariantTy {
+    fn to_value_optional(s: Option<&Self>) -> crate::Value {
+        let mut value = crate::Value::for_value_type::<VariantType>();
+        unsafe {
+            gobject_ffi::g_value_set_boxed(
+                value.to_glib_none_mut().0,
+                s.to_glib_none().0 as *mut _,
+            );
+        }
+
+        value
+    }
 }
 
 impl StaticType for VariantType {
     fn static_type() -> Type {
-        unsafe { from_glib(glib_sys::g_variant_type_get_gtype()) }
+        unsafe { from_glib(ffi::g_variant_type_get_gtype()) }
     }
 }
 
-impl SetValue for VariantType {
-    unsafe fn set_value(value: &mut Value, this: &Self) {
-        gobject_sys::g_value_set_boxed(
-            value.to_glib_none_mut().0,
-            this.to_glib_none().0 as glib_sys::gpointer,
-        )
+#[doc(hidden)]
+impl crate::value::ValueType for VariantType {
+    type Type = VariantType;
+}
+
+#[doc(hidden)]
+unsafe impl<'a> crate::value::FromValue<'a> for VariantType {
+    type Checker = crate::value::GenericValueTypeOrNoneChecker<Self>;
+
+    unsafe fn from_value(value: &'a crate::Value) -> Self {
+        let ptr = gobject_ffi::g_value_get_boxed(value.to_glib_none().0);
+        assert!(!ptr.is_null());
+        from_glib_none(ptr as *const ffi::GVariantType)
     }
 }
 
-impl SetValueOptional for VariantType {
-    unsafe fn set_value_optional(value: &mut Value, this: Option<&Self>) {
-        use std::ptr;
-        let p = match this {
-            Some(ref t) => t.to_glib_none().0 as glib_sys::gpointer,
-            None => ptr::null(),
-        };
-        gobject_sys::g_value_set_boxed(value.to_glib_none_mut().0, p)
+#[doc(hidden)]
+impl crate::value::ToValue for VariantType {
+    fn to_value(&self) -> crate::Value {
+        unsafe {
+            let mut value = crate::Value::from_type(VariantType::static_type());
+            gobject_ffi::g_value_set_boxed(
+                value.to_glib_none_mut().0,
+                ToGlibPtr::<*mut _>::to_glib_none(&self).0 as *mut _,
+            );
+            value
+        }
+    }
+
+    fn value_type(&self) -> crate::Type {
+        VariantType::static_type()
     }
 }
 
-impl<'a> FromValueOptional<'a> for VariantType {
-    unsafe fn from_value_optional(value: &'a Value) -> Option<Self> {
-        Option::<VariantType>::from_glib_none(
-            gobject_sys::g_value_get_boxed(value.to_glib_none().0) as *mut glib_sys::GVariantType,
-        )
+#[doc(hidden)]
+impl crate::value::ToValueOptional for VariantType {
+    fn to_value_optional(s: Option<&Self>) -> crate::Value {
+        let mut value = crate::Value::for_value_type::<Self>();
+        unsafe {
+            gobject_ffi::g_value_set_boxed(
+                value.to_glib_none_mut().0,
+                ToGlibPtr::<*mut _>::to_glib_none(&s).0 as *mut _,
+            );
+        }
+
+        value
     }
 }
 
@@ -320,6 +401,7 @@ impl_eq!(&'a VariantTy, Cow<'b, VariantTy>);
 
 macro_rules! impl_str_eq {
     ($lhs:ty, $rhs: ty) => {
+        #[allow(clippy::redundant_slicing)]
         impl<'a, 'b> PartialEq<$rhs> for $lhs {
             #[inline]
             fn eq(&self, other: &$rhs) -> bool {
@@ -350,12 +432,10 @@ impl Eq for VariantType {}
 #[cfg(test)]
 mod tests {
     use super::*;
-    use glib_sys;
-    use translate::*;
-    use value::ToValue;
+    use crate::ToValue;
 
     unsafe fn equal<T, U>(ptr1: *const T, ptr2: *const U) -> bool {
-        from_glib(glib_sys::g_variant_type_equal(
+        from_glib(ffi::g_variant_type_equal(
             ptr1 as *const _,
             ptr2 as *const _,
         ))
@@ -404,8 +484,8 @@ mod tests {
     fn ne() {
         let ty1 = VariantTy::new("((iii)s)").unwrap();
         let ty2 = VariantTy::new("((iii)o)").unwrap();
-        assert!(ty1 != ty2);
-        assert!(ty1 != "((iii)o)");
+        assert_ne!(ty1, ty2);
+        assert_ne!(ty1, "((iii)o)");
         unsafe {
             assert!(!equal(ty1.as_ptr(), ty2.as_ptr()));
         }
@@ -439,11 +519,18 @@ mod tests {
         assert_eq!(ty1, ty2);
 
         let ty3 = VariantTy::new("*").unwrap();
-        let tyv2 = ty1.to_value();
+        let tyv2 = ty3.to_value();
         let ty4 = tyv2.get::<VariantType>().unwrap();
         assert_eq!(ty3, ty4);
 
-        assert_eq!(VariantTy::static_type(), VariantTy::static_type());
+        let ty5 = VariantTy::ANY;
+        let tyv3 = ty5.to_value();
+        let ty6 = tyv3.get::<VariantType>().unwrap();
+        assert_eq!(ty5, ty6);
     }
 
+    #[test]
+    fn type_() {
+        assert_eq!(VariantTy::static_type(), VariantType::static_type())
+    }
 }
